@@ -3,11 +3,13 @@ import logging
 from queue import SimpleQueue
 import time
 from typing import Any, Callable, Dict, Iterable, List, Optional
+from weakref import ref
 
-from observ import reactive, scheduler, to_raw, watch
+from observ import reactive, scheduler, to_raw
+from observ.watcher import Watcher
 
 from .compare import equivalent_functions
-from .renderers import DictRenderer, Renderer
+from .renderers import Renderer
 from .types import (
     EffectTag,
     EventLoopType,
@@ -19,23 +21,63 @@ from .types import (
 
 logger = logging.getLogger(__name__)
 
+# Node types for which no DOM elements are made
+VIRTUAL_NODE_TYPES = {"template", "slot"}
+
+
+class NotifyChangeWatcher(Watcher):
+    """
+    Custom watcher that simply calls callback function without evaluating
+    the new value of the watched expression.
+    This makes it possible to just signal to Collagraph that a dependency
+    has changed and then Collagraph can decide when to actually evaluate
+    the new value.
+    """
+
+    def update(self):
+        """On update, simply run the callback."""
+        self.callback()
+
+
+def watch(fn, cb, lazy=False, deep=False, **kwargs):
+    """Create custom watcher for the given expression: fn.
+
+    Defaults lazy to False, so that value is evaluated immediately. And deep
+    defaults to True for convenience."""
+    return NotifyChangeWatcher(fn, lazy=lazy, deep=deep, callback=cb)
+
 
 def create_element(type, props=None, *children) -> VNode:
     """Create an element description, based on type, props and (optionally) children"""
-    key = props.pop("key") if props and "key" in props else None
+    key = props.get("key", None) if props is not None else None
+    if len(children) == 1:
+        # If children is 1 dictionary, then that is the slots definition
+        if isinstance(children[0], dict):
+            children = children[0]
+        # If children is 1 callable item, then it becomes the default slot
+        elif callable(children[0]):
+            children = {"default": children[0]}
     return VNode(type, reactive(props or {}), children or tuple(), key)
 
 
+def render_slot(name, props, slots):
+    if name in slots:
+        result = slots[name](props)
+        if isinstance(result, VNode):
+            return [result]
+        return result
+    return ()
+
+
 class Collagraph:
     def __init__(
         self,
-        renderer: Renderer = None,
+        renderer,
         *,
         event_loop_type: EventLoopType = EventLoopType.DEFAULT,
     ):
-        if renderer is None:
-            renderer = DictRenderer()
-        assert isinstance(renderer, Renderer)
+        if not isinstance(renderer, Renderer):
+            raise TypeError(f"Expected a Renderer but got a {type(renderer)}")
         self.renderer = renderer
         self.event_loop_type = event_loop_type
         if self.event_loop_type is EventLoopType.QT:
@@ -84,8 +126,6 @@ class Collagraph:
             deadline: targetted deadline for until when work can be done. If
                 no deadline is given, then it will be set to 16ms from now.
         """
-        logger.debug("Request work")
-
         # current in ns
         if not deadline:
             deadline = time.perf_counter_ns() + 16 * 1000000
@@ -103,7 +143,7 @@ class Collagraph:
                     loop.call_soon(self.work_loop, deadline)
 
                 self._request = start
-            if self.event_loop_type is EventLoopType.QT:  # pragma: no cover
+            if self.event_loop_type is EventLoopType.QT:
                 from PySide6 import QtCore
 
                 self._qt_timer = QtCore.QTimer()
@@ -115,9 +155,12 @@ class Collagraph:
                 def start(deadline):
                     if not self._qt_first_run:
                         self._qt_timer.timeout.disconnect()
+                    else:
                         self._qt_first_run = False
+
+                    weak_self = ref(self)
                     self._qt_timer.timeout.connect(
-                        lambda: self.work_loop(deadline=deadline)
+                        lambda: weak_self() and weak_self().work_loop(deadline=deadline)
                     )
                     self._qt_timer.start()
 
@@ -182,43 +225,63 @@ class Collagraph:
 
         # Attach the component instance to the fiber
         fiber.component = component
-        fiber.component_watcher = watch(
-            lambda: fiber.component.state,
-            lambda: self.state_updated(fiber),
-            deep=True,
-            sync=self.event_loop_type is EventLoopType.SYNC,
-        )
+
+        component._slots = fiber.children if isinstance(fiber.children, dict) else {}
 
         if fiber.alternate:
             fiber.alternate.component = None
-            fiber.alternate.component_watcher = None
+            fiber.watcher = fiber.alternate.watcher
+            fiber.alternate.watcher = None
 
-        children = [component.render()]
-        self.reconcile_children(fiber, children)
+        if fiber.watcher:
+            # Re-evaluate the watcher to get the new value
+            fiber.watcher.evaluate()
+        else:
+            fiber.watcher = watch(
+                component.render,
+                lambda: self.state_updated(fiber),
+            )
+
+        self.reconcile_children(fiber, [fiber.watcher.value])
 
     def update_function_component(self, fiber: Fiber):
-        children = [fiber.type(fiber.props)]
-        self.reconcile_children(fiber, children)
+        if fiber.alternate:
+            fiber.alternate.watcher = None
+
+        def render():
+            if isinstance(fiber.children, dict):
+                return [fiber.type(fiber.props, fiber.children)]
+            return [fiber.type(fiber.props)]
+
+        fiber.watcher = watch(
+            render,
+            lambda: self.state_updated(fiber),
+        )
+
+        self.reconcile_children(fiber, fiber.watcher.value)
 
     def update_host_component(self, fiber: Fiber):
-        # Add dom node
-        if not fiber.dom:
+        # Add dom node, but not for template/slot tags (which are virtual tags)
+        if not fiber.dom and fiber.type not in VIRTUAL_NODE_TYPES:
             fiber.dom = self.create_dom(fiber)
 
+        if fiber.alternate:
+            fiber.alternate.watcher = None
+
+        fiber.watcher = watch(
+            lambda: fiber.props.keys(),
+            lambda: self.state_updated(fiber),
+        )
+
         # Create new fibers
         self.reconcile_children(fiber, fiber.children)
 
     def create_dom(self, fiber: Fiber) -> Any:
         dom = self.renderer.create_element(fiber.type)
-        self.update_dom(fiber, dom, prev_props={}, next_props=fiber.props)
+        self.update_dom_or_component(fiber, dom, prev_props={}, next_props=fiber.props)
         return dom
 
     def state_updated(self, fiber: Fiber):
-        logger.debug(f"state update: {fiber.type}")
-        # Clear the watcher that triggered the update
-        fiber.watcher = None
-        fiber.component_watcher = None
-
         # Request an update to start building/update the wip fiber tree
         self._wip_root = (
             self._current_root and self._current_root.alternate
@@ -244,20 +307,6 @@ class Collagraph:
                 old_fibers.append(sibling)
                 sibling = sibling.sibling
 
-        # Create watcher for the wip_fiber if not already there
-        if wip_fiber.props and not wip_fiber.watcher:
-            wip_fiber.watcher = watch(
-                lambda: wip_fiber.props,
-                lambda: self.state_updated(wip_fiber),
-                deep=True,
-                sync=self.event_loop_type is EventLoopType.SYNC,
-            )
-
-        # Clear the watcher from the old fiber
-        if old_fiber and old_fiber.props:
-            old_fiber.watcher = None
-            old_fiber.component_watcher = None
-
         def matcher(x, y):
             return x.key == y.key
 
@@ -282,13 +331,7 @@ class Collagraph:
 
         # In here, all the 'new' elements are compared to the old/current fiber/state
         prev_sibling = None
-        for idx, (element, old_fiber) in enumerate(
-            zip_longest(elements, ordered_old_fibers + removals)
-        ):
-            # Clear the watcher from the old fiber
-            if old_fiber and old_fiber.props:
-                old_fiber.watcher = None
-
+        for element, old_fiber in zip_longest(elements, ordered_old_fibers + removals):
             new_fiber = None
             same_type = old_fiber and element and element.type == old_fiber.type
 
@@ -371,6 +414,7 @@ class Collagraph:
         component_hooks = SimpleQueue()
         component_hooks.put((self._wip_root, True))
 
+        # parent_component = None
         while not component_hooks.empty():
             # `down` is whether the tree is walked down toward the leaves
             # or up, back towards the root. On the way back, all the children
@@ -449,6 +493,9 @@ class Collagraph:
 
         if fiber.effect_tag == EffectTag.PLACEMENT:
             if fiber.component:
+                self.update_dom_or_component(
+                    fiber, None, prev_props={}, next_props=fiber.props
+                )
                 fiber.mounted = True
             if fiber.dom:
                 self.renderer.insert(fiber.dom, dom_parent, anchor=fiber.anchor)
@@ -461,7 +508,7 @@ class Collagraph:
                         dom_parent,
                         anchor=fiber.anchor,
                     )
-            self.update_dom(
+            self.update_dom_or_component(
                 fiber,
                 fiber.dom,
                 prev_props=fiber.alternate.props_snapshot,
@@ -473,70 +520,44 @@ class Collagraph:
         self._work.put(fiber.child)
         self._work.put(fiber.sibling)
 
-    def update_dom(self, fiber: Fiber, dom: Any, prev_props: Dict, next_props: Dict):
-        def is_event(key):
-            return key.startswith("on_")
-
-        def key_to_event(key):
-            # Events start with `on_`
-            return key.lower()[3:]
-
-        def is_property(key):
-            return not is_event(key)
-
-        def is_new(val, other, key):
-            return val != other.get(key)
-
-        def is_equivalent_event_handler(val, other, key):
-            alt = other.get(key)
-            return equivalent_functions(val, alt)
+    def update_dom_or_component(
+        self, fiber: Fiber, dom: Any, prev_props: Dict, next_props: Dict
+    ):
+        if not dom and not fiber.component:
+            return
 
         events_to_remove = {}
-        # Remove old event listeners
-        for name, val in prev_props.items():
-            if not is_event(name):
-                continue
-            if name in next_props and is_equivalent_event_handler(
-                val, next_props, name
-            ):
-                continue
-
-            event_type = key_to_event(name)
-            events_to_remove[event_type] = val
-
-        # Remove old properties
         attrs_to_remove = {}
-        for key in prev_props:
-            # Is key an actual property?
-            if not is_property(key):
-                continue
-            # Is key gone?
-            if key in next_props:
-                continue
-
-            attrs_to_remove[key] = prev_props[key]
-
         attrs_to_update = {}
-        # Set new or changed properties
-        for key, val in next_props.items():
-            if not is_property(key):
-                continue
-            # Is key new or changed?
-            if not is_new(val, prev_props, key):
-                continue
-
-            attrs_to_update[key] = val
-
         events_to_add = {}
-        # Add new event listeners
-        for name in next_props:
-            if not is_event(name):
-                continue
-            if is_equivalent_event_handler(prev_props.get(name), next_props, name):
-                continue
+        equivalent_event_handlers = set()
 
-            event_type = key_to_event(name)
-            events_to_add[event_type] = next_props[name]
+        for key, val in prev_props.items():
+            if is_event(key):
+                if key in next_props and is_equivalent_event_handler(
+                    val, next_props, key
+                ):
+                    equivalent_event_handlers.add(key)
+                    continue
+
+                event_type = key_to_event(key)
+                events_to_remove[event_type] = val
+            elif dom:  # Don't track prop changes for non-dom fibers
+                # Is key gone?
+                if key not in next_props:
+                    attrs_to_remove[key] = prev_props[key]
+
+        for key, val in next_props.items():
+            if is_event(key):
+                if key in equivalent_event_handlers:
+                    continue
+
+                event_type = key_to_event(key)
+                events_to_add[event_type] = next_props[key]
+            elif dom:  # Don't track prop changes for non-dom fibers
+                # Is key new or changed?
+                if is_new(val, prev_props, key):
+                    attrs_to_update[key] = val
 
         if dom:
             for event_type, val in events_to_remove.items():
@@ -551,17 +572,42 @@ class Collagraph:
             for event_type, val in events_to_add.items():
                 self.renderer.add_event_listener(dom, event_type, val)
 
-        # Climb up the tree to find first component to mark as updated
-        if events_to_remove or events_to_add or attrs_to_remove or attrs_to_update:
-            parent = fiber.parent
-            component = None
-            while parent:
-                component = parent.component
-                if component:
-                    # Mark the fiber as updated
-                    parent.updated = True
-                    break
-                parent = parent.parent
+            # Climb up the tree to find first component to mark as updated
+            if events_to_remove or events_to_add or attrs_to_remove or attrs_to_update:
+                parent = fiber.parent
+                component = None
+                while parent:
+                    component = parent.component
+                    if component:
+                        # Mark the fiber as updated
+                        parent.updated = True
+                        break
+                    parent = parent.parent
+
+        if fiber.component:
+            for event_type, val in events_to_remove.items():
+                fiber.component.remove_event_handler(event_type, val)
+
+            for event_type, val in events_to_add.items():
+                fiber.component.add_event_handler(event_type, val)
+
+
+def is_event(key):
+    return key.startswith("on_")
+
+
+def key_to_event(key):
+    # Events start with `on_`
+    return key[3:].lower()
+
+
+def is_new(val, other, key):
+    return val != other.get(key)
+
+
+def is_equivalent_event_handler(val, other, key):
+    alt = other.get(key)
+    return equivalent_functions(val, alt)
 
 
 def indexOf(items: Iterable, match: Callable, *args):
@@ -644,7 +690,7 @@ def create_ops(current, future):
     wip = current.copy()
 
     # First figure out all the deletions that need to take place
-    for idx, old in enumerate(current):
+    for old in current:
         if old not in future:
             ops.append(create_operation(OpType.DEL, value=old))
             apply_op(ops[-1], wip)
